@@ -3,11 +3,10 @@ use std::collections::HashMap;
 use std::io::BufReader;
 
 use super::*;
-use crate::builtin;
+use crate::builtin::Builtin;
 use crate::error::Error;
 use crate::reader::Reader;
-use crate::token;
-use crate::value::Value;
+use crate::value::{RefValue, Token};
 use crate::vm::*;
 
 /** Compiler symbolic scope.
@@ -32,7 +31,7 @@ pub(super) enum Scope {
         usage_start: usize, // Begin of usages to resolve until when scope is closed
         constants: HashMap<String, ImlValue>, // Constants symbol table
     },
-    ImlLoop, // loop level (allows use of break & continue)
+    Loop, // loop level (allows use of break & continue)
 }
 
 /** Tokay compiler instance
@@ -113,9 +112,9 @@ impl Compiler {
     }
 
     /// Shortcut to compile a Tokay program from a &str.
-    pub fn compile_str(&mut self, src: &'static str) -> Result<Program, Vec<Error>> {
+    pub fn compile_str(&mut self, src: &str) -> Result<Program, Vec<Error>> {
         self.compile(Reader::new(Box::new(BufReader::new(std::io::Cursor::new(
-            src,
+            src.to_owned(),
         )))))
     }
 
@@ -161,7 +160,9 @@ impl Compiler {
 
         // Obtain intermediate values collected during compilation
         let values = if self.interactive {
-            self.values.clone()
+            let values = self.values.clone();
+            self.values.pop(); // Remove last __main__ from interactive compiler.
+            values
         } else {
             self.values.drain(..).collect()
         };
@@ -242,7 +243,9 @@ impl Compiler {
             values
                 .into_iter()
                 .map(|value| match value {
-                    ImlValue::Parselet(parselet) => parselet.borrow().into_parselet().into_value(),
+                    ImlValue::Parselet(parselet) => {
+                        RefValue::from(parselet.borrow().into_parselet())
+                    }
                     ImlValue::Value(value) => value,
                 })
                 .collect(),
@@ -301,7 +304,7 @@ impl Compiler {
 
     /// Push a loop scope
     pub(super) fn push_loop(&mut self) {
-        self.scopes.insert(0, Scope::ImlLoop);
+        self.scopes.insert(0, Scope::Loop);
     }
 
     /// Resolves and drops a parselet scope and creates a new parselet from it.
@@ -352,6 +355,7 @@ impl Compiler {
             };
 
             if self.scopes.len() == 0 && self.interactive {
+                *consuming = false;
                 self.scopes.push(scope);
             }
 
@@ -370,7 +374,7 @@ impl Compiler {
 
     /// Drops a loop scope.
     pub(super) fn pop_loop(&mut self) {
-        assert!(self.scopes.len() > 0 && matches!(self.scopes[0], Scope::ImlLoop));
+        assert!(self.scopes.len() > 0 && matches!(self.scopes[0], Scope::Loop));
         self.scopes.remove(0);
     }
 
@@ -391,7 +395,7 @@ impl Compiler {
         for i in 0..self.scopes.len() {
             match &self.scopes[i] {
                 Scope::Parselet { .. } => return false,
-                Scope::ImlLoop => return true,
+                Scope::Loop => return true,
                 _ => {}
             }
         }
@@ -537,21 +541,24 @@ impl Compiler {
         }
 
         // When not found, check for a builtin function
-        if let Some(builtin) = builtin::get(name) {
-            return Some(Value::Builtin(builtin).into());
+        if let Some(builtin) = Builtin::get(name) {
+            return Some(RefValue::from(builtin).into()); // fixme: Makes a Value into a RefValue into a Value...
         }
 
         // Builtin constants are defined on demand as fallback
         if name == "_" || name == "__" {
             // Fallback for "_" defines parselet `_ : Whitespace?`
-            self.set_constant("_", token::get("Whitespaces").unwrap().into_value().into());
+            self.set_constant(
+                "_",
+                RefValue::from(Token::builtin("Whitespaces").unwrap()).into(),
+            );
 
             return Some(self.get_constant(name).unwrap());
         }
 
         // Check for built-in token
-        if let Some(value) = token::get(name) {
-            return Some(value.into_value().into());
+        if let Some(value) = Token::builtin(name) {
+            return Some(RefValue::from(value).into());
         }
 
         None
@@ -575,3 +582,124 @@ impl Compiler {
         self.values.len() - 1
     }
 }
+
+#[test]
+fn test_whitespace() {
+    // Builtin whitespace handling
+    let abc = "abc   \tdef  abcabc= ghi abcdef";
+
+    assert_eq!(
+        crate::run("Word _; ", abc),
+        Ok(Some(crate::value![[
+            "abc", "def", "abcabc", "ghi", "abcdef"
+        ]]))
+    );
+
+    assert_eq!(
+        crate::run("Word __; ", abc),
+        Ok(Some(crate::value![["abc", "def", "ghi"]]))
+    );
+}
+
+#[test]
+// Testing several special parsing constructs and error reporting
+fn test_error_reporting() {
+    // Test for programs which consist just of one comment
+    assert_eq!(crate::run("#tralala", ""), Ok(None));
+
+    // Test for whitespace
+    assert_eq!(
+        crate::run("#normal comment\n#\n\t123", ""),
+        Ok(Some(crate::value!(123)))
+    );
+
+    // Test for invalid input when EOF is expected
+    assert_eq!(
+        crate::run("{}}", ""),
+        Err("Line 1, column 3: Parse error, expecting end-of-file".to_string())
+    );
+
+    // Test on unclosed sequences `(1 `
+    assert_eq!(
+        crate::run("(1", ""),
+        Err("Line 1, column 3: Expecting \")\"".to_string())
+    );
+
+    assert_eq!(
+        crate::run("(a => 1, b => 2", ""),
+        Err("Line 1, column 16: Expecting \")\"".to_string())
+    );
+
+    // Test empty sequence
+    assert_eq!(crate::run("()", ""), Ok(None));
+
+    // Tests on filled and empty blocks and empty blocks
+    assert_eq!(
+        crate::run(
+            "
+            a = {}
+            b = {
+            }
+            c = {
+                1
+                2
+                3
+            }
+
+            a b c
+            ",
+            ""
+        ),
+        Ok(Some(crate::value!(3)))
+    );
+}
+
+#[test]
+// Tests for correct identifier names for various value types
+fn test_identifier_naming() {
+    crate::test::testcase("tests/err_compiler_identifier_names.tok");
+}
+
+#[test]
+// Tests for compiler string, match and ccl escaping
+fn test_unescaping() {
+    assert_eq!(
+        crate::run(
+            "\"test\\\\yes\n\\xCA\\xFE\t\\100\\x5F\\u20ac\\U0001F98E\"",
+            ""
+        ),
+        Ok(Some(crate::value!("test\\yes\nÊþ\t@_€🦎")))
+    );
+
+    assert_eq!(
+        crate::run(
+            "'hello\\nworld'", // double \ quotation required
+            "hello\nworld"
+        ),
+        Ok(Some(crate::value!("hello\nworld")))
+    );
+
+    assert_eq!(
+        crate::run(
+            "[0-9\\u20ac]+", // double \ quotation required
+            "12345€ €12345"
+        ),
+        Ok(Some(crate::value!(["12345€", "€12345"])))
+    );
+
+    assert_eq!(
+        crate::run(
+            "'hello\\nworld'", // double \ quotation required
+            "hello\nworld"
+        ),
+        Ok(Some(crate::value!("hello\nworld")))
+    );
+
+    assert_eq!(
+        crate::run(
+            "[0-9\\u20ac]+", // double \ quotation required
+            "12345€ €12345"
+        ),
+        Ok(Some(crate::value!(["12345€", "€12345"])))
+    );
+}
