@@ -1,16 +1,18 @@
-//! Parser and grammar for Tokay, implemented using Tokay itself.
+/*! Parser and grammar for Tokay, implemented using Tokay itself.
+
+This module implements a Tokay parser in Tokay itself, using the compiler macros from the macros-module.
+This is the general place to change syntax and modify the design of the abstract syntax tree.
+
+There is also a grammar in examples/tokay.tok, which is this grammar expressed in Tokay.
+When changing things here, do it there as well as this might become the reference parser someday.
+*/
+use charclass::charclass;
 
 use super::*;
 use crate::error::Error;
 use crate::reader::Reader;
-use crate::token::Token;
-use crate::value::Value;
-use crate::{ccl, tokay, value};
-
-/**
-Implements a Tokay parser in Tokay itself, using the compiler macros from the macros-module.
-This is the general place to change syntax and modify the design of the abstract syntax tree.
-*/
+use crate::tokay;
+use crate::value::{Dict, RefValue, Str, Token};
 
 pub struct Parser(Program);
 
@@ -22,31 +24,31 @@ impl Parser {
         // Whitespace & EOL
 
         (_ = {  // true whitespace
-            [" "],
-            ["#", (token (Token::Chars(ccl!['\n'].negate())))],
+            [(token (Token::Chars(charclass![' ', '\t'])))],
+            ["#", (opt (token (Token::Chars(charclass!['\n'].negate()))))],
             ["\\", "\n"]
         }),
 
         (___ = {  // check for non-trailing identifier
-            [(peek (not (token (Token::Char(ccl!['A' => 'Z', 'a' => 'z'] + ccl!['_']))))), _]
+            [(peek (not (token (Token::Char(charclass!['A' => 'Z', 'a' => 'z'] + charclass!['_']))))), _]
         }),
 
         (T_EOL = {  // end-of-line
-            ["\n", _, (Op::Skip)],
-            [";", _, (Op::Skip)],
-            ["|", _, (Op::Skip)],
-            [(token (Token::EOF)), (Op::Skip)],
-            [(peek "}"), (Op::Skip)]
+            ["\n", _, (Op::Skip)],  // unix/linux
+            ["\r", (opt "\n"), _, (Op::Skip)],  // classic mac & windows
+            [";", _, (Op::Skip)],  // allows for multiple lines in one row
+            [(token (Token::EOF)), (Op::Skip)],  // Input file may end without a new line
+            [(peek "}"), (Op::Skip)]  // peek for '}' to allow for blocks in one line.
         }),
 
         // Escape sequences
 
         (T_OctDigit = {  // T_OctDigit is used by T_EscapeSequence
-            (token (Token::Char(ccl!['0' => '7'])))
+            (token (Token::Char(charclass!['0' => '7'])))
         }),
 
         (T_HexDigit = {    // T_HexDigit is used by T_EscapeSequence
-            (token (Token::Char(ccl!['0' => '9', 'A' => 'F', 'a' => 'f'])))
+            (token (Token::Char(charclass!['0' => '9', 'A' => 'F', 'a' => 'f'])))
         }),
 
         (T_EscapeSequence = {   // Parsing escape sequences
@@ -69,28 +71,24 @@ impl Parser {
             Any
         }),
 
-        // Prime Tokens (might probably be replaced by something native, pluggable one)
+        // Prime Tokens
 
         (T_Identifier = {  // any identifier
-            [
-                (token (Token::Char(ccl!['A' => 'Z', 'a' => 'z'] + ccl!['_']))),
-                (opt (token (Token::Chars(ccl!['A' => 'Z', 'a' => 'z', '0' => '9'] + ccl!['_'])))),
-                (call ast[(value "identifier"), (Op::LoadFastCapture(0))])
-            ]
+            (call ast[(value "identifier"), (call Ident[])])
         }),
 
         (T_Consumable = {  // consumable identifier
             [
-                (token (Token::Char(ccl!['A' => 'Z'] + ccl!['_']))),
-                (opt (token (Token::Chars(ccl!['A' => 'Z', 'a' => 'z', '0' => '9'] + ccl!['_'])))),
+                (token (Token::Char(charclass!['A' => 'Z'] + charclass!['_']))),
+                (opt (token (Token::Chars(charclass!['A' => 'Z', 'a' => 'z', '0' => '9'] + charclass!['_'])))),
                 (call ast[(value "identifier"), (Op::LoadFastCapture(0))])
             ]
         }),
 
         (T_Alias = {  // T_Alias is an identifier treated as string value
             [
-                (token (Token::Char(ccl!['A' => 'Z', 'a' => 'z'] + ccl!['_']))),
-                (opt (token (Token::Chars(ccl!['A' => 'Z', 'a' => 'z', '0' => '9'] + ccl!['_'])))),
+                (token (Token::Char(charclass!['A' => 'Z', 'a' => 'z'] + charclass!['_']))),
+                (opt (token (Token::Chars(charclass!['A' => 'Z', 'a' => 'z', '0' => '9'] + charclass!['_'])))),
                 (call ast[(value "value_string"), (Op::LoadFastCapture(0))])
             ]
         }),
@@ -100,7 +98,7 @@ impl Parser {
                 "\"",  // a string
                 (kle {
                     ["\\", T_EscapeSequence],
-                    [(token (Token::Chars(ccl!['\\', '\"'].negate())))],
+                    [(token (Token::Chars(charclass!['\\', '\"'].negate())))],
                     [EOF, (call error[(value "Unclosed string, expecting '\"'")])]
                 }),
                 (call str_join[(value ""), (Op::LoadFastCapture(2))]),
@@ -113,7 +111,7 @@ impl Parser {
                 "\'",  // a touch
                 (kle {
                     ["\\", T_EscapeSequence],
-                    [(token (Token::Chars(ccl!['\\', '\''].negate())))],
+                    [(token (Token::Chars(charclass!['\\', '\''].negate())))],
                     [EOF, (call error[(value "Unclosed match, expecting '\''")])]
                 }),
                 (call str_join[(value ""), (Op::LoadFastCapture(2))]),
@@ -122,29 +120,32 @@ impl Parser {
         }),
 
         (T_Integer = {
-            // todo: implement as built-in Parselet
-            [(token (Token::Chars(ccl!['0' => '9']))), (call ast[(value "value_integer")])]
+            [(call ast[(value "value_integer"), (call Int[])])]
         }),
 
         (T_Float = {
-            // todo: implement as built-in Parselet
-            [(token (Token::Chars(ccl!['0' => '9']))), ".", (opt (token (Token::Chars(ccl!['0' => '9'])))),
-                (call ast[(value "value_float"), (Op::LoadFastCapture(0))])],
-            [(opt (token (Token::Chars(ccl!['0' => '9'])))), ".", (token (Token::Chars(ccl!['0' => '9']))),
-                (call ast[(value "value_float"), (Op::LoadFastCapture(0))])]
+            [(call ast[(value "value_float"), (call Float[])])]
         }),
 
         // Character classes
 
         (CclChar = {
             ["\\", T_EscapeSequence],
-            (token (Token::Char(ccl![']'].negate()))),
+            (token (Token::Char(charclass![']'].negate()))),
             [EOF, (call error[(value "Unclosed character-class, expecting ']'")])]
         }),
 
         (CclRange = {
             [CclChar, "-", CclChar,
-                (call ast[(value "range"), [(Op::LoadFastCapture(1)), (Op::LoadFastCapture(3)), (Op::Add)]])],
+                (call ast[
+                    (value "range"),
+                    [
+                        (Op::LoadFastCapture(1)),
+                        (Op::LoadFastCapture(3)),
+                        (Op::BinaryOp("add"))
+                    ]
+                ])
+            ],
             [CclChar, (call ast[(value "char")])]
         }),
 
@@ -159,6 +160,10 @@ impl Parser {
             ["[", _, Expression, "]", _, (call ast[(value "index")])]
         }),
 
+        (Attribute = {
+            [".", _, (expect T_Alias), (call ast[(value "attribute")])]
+        }),
+
         (Capture = {
             ["$", T_Alias, _, (call ast[(value "capture_alias")])],
             ["$", T_Integer, _, (call ast[(value "capture_index")])],
@@ -172,9 +177,10 @@ impl Parser {
         }),
 
         (Lvalue = {
-            //[Lvalue, ".", _, T_Alias, (call ast[(value "attribute")])],
-            //[Lvalue, (pos Subscript)],
-            [Variable, (call ast[(value "lvalue")])]
+            [Variable, _, (kle {
+                // Attribute,  // Attribute assignment not required for now.
+                Subscript
+            }), (call ast[(value "lvalue")])]
         }),
 
         (Load = {
@@ -218,7 +224,8 @@ impl Parser {
         }),
 
         (Collection = {
-            ["(", _, (kle [T_EOL, _]), (pos [Expression, (opt [",", _]), (kle [T_EOL, _])]), ")", // no expect ")" here!
+            ["(", _, (kle [T_EOL, _]), ")", (call ast[(value "value_void")])],
+            ["(", _, (kle [T_EOL, _]), (expect (pos [Expression, (opt [",", _]), (kle [T_EOL, _])])), ")", // no expect ")" here!
                 (call ast[(value "sequence")])],
             ["(", _, (kle [T_EOL, _]), (pos [CollectionItem, (opt [",", _]), (kle [T_EOL, _])]), (expect ")"),
                 (call ast[(value "sequence")])]
@@ -283,10 +290,12 @@ impl Parser {
         }),
 
         (Rvalue = {
-            [Rvalue, ".", _, T_Alias, (call ast[(value "attribute")])],
-            [Rvalue, (pos Subscript), (call ast[(value "rvalue")])],
-            [Rvalue, "(", _, (kle [T_EOL, _]), (opt CallParameters), (expect ")"), _,
+            [Rvalue, "(", _, (kle [T_EOL, _]), (opt CallParameters), (expect ")"),
                 (call ast[(value "call")])],
+            [Rvalue, (kle {
+                Attribute,
+                Subscript
+            }), (call ast[(value "rvalue")])],
             Atomic
         }),
 
@@ -311,12 +320,12 @@ impl Parser {
         }),
 
         (Compare = {
-            [Compare, "==", _, (expect AddSub), (call ast[(value "op_compare_equal")])],
-            [Compare, "!=", _, (expect AddSub), (call ast[(value "op_compare_unequal")])],
-            [Compare, "<=", _, (expect AddSub), (call ast[(value "op_compare_lowerequal")])],
-            [Compare, ">=", _, (expect AddSub), (call ast[(value "op_compare_greaterequal")])],
-            [Compare, "<", _, (expect AddSub), (call ast[(value "op_compare_lower")])],
-            [Compare, ">", _, (expect AddSub), (call ast[(value "op_compare_greater")])],
+            [Compare, "==", _, (expect AddSub), (call ast[(value "op_compare_eq")])],
+            [Compare, "!=", _, (expect AddSub), (call ast[(value "op_compare_neq")])],
+            [Compare, "<=", _, (expect AddSub), (call ast[(value "op_compare_lteq")])],
+            [Compare, ">=", _, (expect AddSub), (call ast[(value "op_compare_gteq")])],
+            [Compare, "<", _, (expect AddSub), (call ast[(value "op_compare_lt")])],
+            [Compare, ">", _, (expect AddSub), (call ast[(value "op_compare_gt")])],
             AddSub
         }),
 
@@ -399,22 +408,27 @@ impl Parser {
             Statement
         }),
 
-        (ImlSequence = {
+        (Sequence = {
             [(pos [SequenceItem, (opt [",", _])]), (call ast[(value "sequence")])]
         }),
 
+        (Sequences = {
+            [Sequence, (pos ["|", _, Sequence]), (call ast[(value "block")])],
+            Sequence
+        }),
+
         (SequenceOrExpression = {
             [Expression, (peek T_EOL)],
-            ImlSequence
+            Sequences
         }),
 
         (Instruction = {
-            ["begin", ___, ImlSequence, (expect T_EOL), (call ast[(value "begin")])],
-            ["end", ___, ImlSequence, (expect T_EOL), (call ast[(value "end")])],
+            ["begin", ___, Sequence, (expect T_EOL), (call ast[(value "begin")])],
+            ["end", ___, Sequence, (expect T_EOL), (call ast[(value "end")])],
 
             [T_Identifier, _, ":", _, (expect SequenceOrExpression), (expect T_EOL),
                 (call ast[(value "constant")])],
-            ImlSequence,
+            Sequences,
             [T_EOL, (Op::Skip)]
         }),
 
@@ -424,7 +438,7 @@ impl Parser {
                 (call error[(value "Parse error, unexpected token"), (value true)])]
         }),
 
-        [_, Tokay,
+        [_, (opt Tokay),
             (expect (token (Token::EOF)), "Parse error, expecting end-of-file"),
             (call ast[(value "main")])]
 
@@ -438,8 +452,7 @@ impl Parser {
         }),
 
         (T_Integer = {
-            // todo: implement as built-in Parselet
-            [(token (Token::Chars(ccl!['0' => '9']))), (call ast[(value "value_integer")])]
+            [(token (Token::Chars(charclass!['0' => '9']))), (call ast[(value "value_integer")])]
         }),
 
         (Instruction = {
@@ -457,11 +470,22 @@ impl Parser {
             (call ast[(value "main")])]
         */
 
+        /*
+        (T_Float = {
+            [(token (Token::Chars(charclass!['0' => '9']))), ".", (opt (token (Token::Chars(charclass!['0' => '9'])))),
+                (call ast[(value "value_float"), (Op::LoadFastCapture(0))])],
+            [(opt (token (Token::Chars(charclass!['0' => '9'])))), ".", (token (Token::Chars(charclass!['0' => '9']))),
+                (call ast[(value "value_float"), (Op::LoadFastCapture(0))])]
+        }),
+
+        T_Float
+        */
+
         // ----------------------------------------------------------------------------
                     }))
     }
 
-    pub fn parse(&self, mut reader: Reader) -> Result<Value, Error> {
+    pub fn parse(&self, mut reader: Reader) -> Result<RefValue, Error> {
         //self.0.dump();
         let mut runtime = Runtime::new(&self.0, &mut reader);
 
@@ -473,13 +497,13 @@ impl Parser {
 
         match self.0.run(&mut runtime) {
             Ok(Some(ast)) => {
-                if ast.get_dict().is_some() {
+                if ast.borrow().object::<Dict>().is_some() {
                     Ok(ast)
                 } else {
                     Err(Error::new(None, "Parse error".to_string()))
                 }
             }
-            Ok(None) => Ok(Value::Void),
+            Ok(None) => Ok(crate::value!(void)),
             Err(error) => Err(error),
         }
     }
@@ -488,7 +512,7 @@ impl Parser {
 fn code_to_char(context: &mut Context, skip: u8, base: u32) -> Result<Accept, Reject> {
     let value = context.get_capture(0).unwrap();
     let value = value.borrow();
-    let slice = &value.get_string().unwrap()[skip as usize..];
+    let slice = &value.object::<Str>().unwrap().as_str()[skip as usize..];
 
     let code = if slice.len() <= 2 {
         u8::from_str_radix(slice, base).unwrap_or_default() as char
@@ -497,6 +521,98 @@ fn code_to_char(context: &mut Context, skip: u8, base: u32) -> Result<Accept, Re
             .unwrap_or_default()
     };
 
-    let value = Value::String(format!("{}", code)).into_refvalue();
-    Ok(Accept::Return(Some(value)))
+    Ok(Accept::Return(Some(RefValue::from(format!("{}", code)))))
+}
+
+#[test]
+// EOL
+fn parser_eol() {
+    for eol in ["\n", "\r", "\r\n", ";"] {
+        let tok = format!("a = 1{}a + 2", eol);
+        println!("EOL test {:?}", tok);
+        assert_eq!(crate::run(&tok, ""), Ok(Some(crate::value!(3))));
+    }
+}
+
+// Tests for parsing and packrat features ---------------------------------------------------------
+
+/*
+    Below are some tests that provide indirect left-recursion.
+
+    They currently don't work properly due to the following reason:
+    For indirect left-recursion in packrat parsing, one rule in the
+    grammar's graph must be declared as "leading", so that subsequent,
+    even left-recursive parselets are considered as not left-recursive.
+
+
+    An implementation of an solution for this issue can be found in
+    the pegen parser generator from Python:
+
+    https://github.com/python/cpython/blob/main/Tools/peg_generator/pegen/parser_generator.py
+
+    Tokay won't take care of this right now as it is an edge-case
+    and also more complex, as Tokay does not directly implements a
+    grammar.
+*/
+
+#[test]
+fn parser_indirectleftrec() {
+    let program = tokay!({
+        (X = {
+            [Y, (MATCH "c")]
+        }),
+        (Y = {
+            [Z, (MATCH "b")]
+            //Void
+        }),
+        (Z = {
+            X,
+            Y,
+            (MATCH "a")
+        }),
+        Z
+    });
+
+    println!("{:#?}", program.run_from_str("aaabc"));
+}
+
+#[test]
+fn parser_leftrec() {
+    /*
+    let program = tokay!({
+        (X = {
+            [X, (MATCH "b")],
+            (MATCH "a")
+        }),
+
+        X
+    });
+    */
+
+    let program = tokay!({
+        (Y = {
+            X,
+            (MATCH "a")
+        }),
+        (X = {
+            [Y, (MATCH "b")]
+        }),
+        X
+    });
+
+    /*
+    let program = tokay!({
+        (Factor = {
+            ["(", (pos [Expression]), ")"],
+            (token (Token::Chars(charclass!['0'..='9'])))
+        }),
+        (Expression = {
+            [Expression, "+", Expression],
+            Factor
+        }),
+        Expression
+    });
+    */
+
+    println!("{:#?}", program.run_from_str("abb"));
 }
