@@ -1,11 +1,11 @@
-//! Universal interface to let Tokay read input from anywhere
+//! Universal low-level interface to let Tokay read input from different sources.
 
 use std::io::prelude::*;
 
 /// Position inside a reader, with row and column counting.
 #[derive(Debug, Clone, Copy, PartialEq)]
 pub struct Offset {
-    // fixme: Hold source filename information as well in the future?
+    // todo: Hold source filename information as well in the future?
     pub offset: usize,
     pub row: u32,
     pub col: u32,
@@ -24,7 +24,7 @@ pub struct Reader {
 impl Reader {
     /// Creates a new reader on buffer read.
     pub fn new(reader: Box<dyn BufRead>) -> Self {
-        let mut ret = Self {
+        Self {
             reader,
             buffer: String::with_capacity(1024), //fixme: Modifyable capacity?
             offset: Offset {
@@ -33,10 +33,7 @@ impl Reader {
                 col: 1,
             },
             eof: false,
-        };
-
-        ret.peek(); // Peek one character to find out if we're immediately EOF
-        ret
+        }
     }
 
     /// Internal function for reading a line.
@@ -95,10 +92,14 @@ impl Reader {
         self.offset
     }
 
-    pub fn eof(&self) -> bool {
+    pub fn eof(&mut self) -> bool {
         if self.buffer[self.offset.offset..].chars().next().is_some() {
             false
         } else {
+            if !self.eof {
+                self.peek();
+            }
+
             self.eof
         }
     }
@@ -131,8 +132,9 @@ impl Reader {
         println!("{:?}", &self.buffer[start..self.offset.offset])
     }
 
-    pub fn extract(&self, range: &Range) -> String {
-        self.buffer[range.start..range.end].to_string()
+    /// Get slice from range
+    pub fn get(&self, range: &Range) -> &str {
+        &self.buffer[range.start..range.end]
     }
 
     /// Commits current input buffer and removes cached content
@@ -140,4 +142,34 @@ impl Reader {
         self.buffer.drain(0..self.offset.offset);
         self.offset.offset = 0;
     }
+
+    /// Take one character accepted by callback
+    pub fn take<F>(&mut self, accept: F) -> Option<char>
+    where
+        F: Fn(char) -> bool,
+    {
+        if let Some(ch) = self.peek() {
+            if accept(ch) {
+                return Some(self.next().unwrap());
+            }
+        }
+
+        None
+    }
+
+    /// Read while conditional callback accepts characters
+    pub fn span<F>(&mut self, accept: F) -> Option<&str>
+    where
+        F: Fn(char) -> bool + Copy,
+    {
+        let start = self.offset.offset;
+
+        while self.take(accept).is_some() {}
+
+        if start < self.offset.offset {
+            Some(&self.buffer[start..self.offset.offset])
+        } else {
+            None
+        }
+    }
 }
