@@ -1,8 +1,10 @@
+//! Parselet object represents a callable, user-defined function.
+
 use std::cell::RefCell;
 use std::rc::Rc;
 
-use super::*;
-use crate::compiler::ast;
+use super::{BoxedObject, Dict, List, Object, RefValue};
+
 use crate::error::Error;
 use crate::vm::*;
 
@@ -21,14 +23,14 @@ the generated parse tree automatically until no more input can be consumed.
 
 #[derive(Debug)]
 pub struct Parselet {
-    pub(crate) name: Option<String>, // Parselet's name from source (for debugging)
+    pub(crate) name: String, // Parselet's name from source (for debugging)
     pub(crate) consuming: Option<bool>, // Indicator for consuming & left-recursion
-    pub(crate) severity: u8,         // Capture push severity
+    pub(crate) severity: u8, // Capture push severity
     signature: Vec<(String, Option<usize>)>, // Argument signature with default arguments
-    pub(crate) locals: usize,        // Number of local variables present
-    begin: Vec<Op>,                  // Begin-operations
-    end: Vec<Op>,                    // End-operations
-    body: Vec<Op>,                   // Operations
+    pub(crate) locals: usize, // Number of local variables present
+    begin: Vec<Op>,          // Begin-operations
+    end: Vec<Op>,            // End-operations
+    body: Vec<Op>,           // Operations
 }
 
 impl Parselet {
@@ -48,8 +50,8 @@ impl Parselet {
             "signature may not be longer than locals..."
         );
 
-        Self {
-            name,
+        let mut ret = Self {
+            name: name.unwrap_or(String::new()),
             consuming,
             severity,
             signature,
@@ -57,26 +59,19 @@ impl Parselet {
             begin,
             end,
             body,
+        };
+
+        if ret.name.is_empty() {
+            ret.name = format!("parselet_{:x}", &ret as *const Parselet as usize);
         }
-    }
 
-    /// Turns parselet into a Value
-    pub fn into_value(self) -> Value {
-        Value::Parselet(Rc::new(RefCell::new(self)))
-    }
-
-    // Checks if parselet is callable with or without arguments
-    pub(crate) fn is_callable(&self, with_arguments: bool) -> bool {
-        // Either without arguments and signature is empty or all arguments have default values
-        (!with_arguments && (self.signature.len() == 0 || self.signature.iter().all(|arg| arg.1.is_some())))
-        // or with arguments and signature exists
-            || (with_arguments && self.signature.len() > 0)
+        ret
     }
 
     fn _run(&self, context: &mut Context, main: bool) -> Result<Accept, Reject> {
         // Initialize parselet execution loop
         let mut first = self.begin.len() > 0;
-        let mut results = Vec::new();
+        let mut results = List::new();
         let mut state = if self.begin.len() == 0 {
             None
         } else {
@@ -86,17 +81,14 @@ impl Parselet {
         // Debugging
         let mut debug = context.runtime.debug;
         if debug < 3 {
-            if let (Ok(inspect), Some(name)) = (std::env::var("TOKAY_INSPECT"), self.name.as_ref())
-            {
-                if inspect.find(name).is_some() {
+            if let Ok(inspect) = std::env::var("TOKAY_INSPECT") {
+                if inspect.find(&self.name).is_some() {
                     debug = 6;
                 }
             }
         }
 
         let result = loop {
-            let reader_start = context.runtime.reader.tell();
-
             let ops = match state {
                 // begin
                 Some(true) => &self.begin,
@@ -110,9 +102,23 @@ impl Parselet {
 
             let mut result = Op::execute(ops, context, debug);
 
-            // if main {
-            //     println!("state = {:?} result = {:?}", state, result);
-            // }
+            // Either take $0 if it was set to a value, or
+            // use any last remaining value as result,
+            // if available
+            if let Ok(Accept::Next) = result {
+                let dollar0 = &mut context.runtime.stack[context.capture_start - 1];
+
+                if let Capture::Value(value, ..) = dollar0 {
+                    result = Ok(Accept::Push(value.clone().into()));
+                } else if context.runtime.stack.len() > context.capture_start {
+                    result = Ok(Accept::Push(context.runtime.stack.pop().unwrap()));
+                }
+            }
+
+            // Debug
+            if context.runtime.debug > 2 {
+                context.debug(&format!("returns {:?}", state));
+            }
 
             /*
                 In case this is the main parselet, try matching main as much
@@ -126,9 +132,9 @@ impl Parselet {
                     Ok(Accept::Return(value)) => Ok(Accept::Repeat(value)),
 
                     Ok(Accept::Push(capture)) => Ok(Accept::Repeat(match capture {
-                        Capture::Range(range, ..) => Some(
-                            Value::String(context.runtime.reader.extract(&range)).into_refvalue(),
-                        ),
+                        Capture::Range(range, ..) => {
+                            Some(RefValue::from(context.runtime.reader.get(&range)))
+                        }
                         Capture::Value(value, ..) => Some(value),
                         _ => None,
                     })),
@@ -182,11 +188,19 @@ impl Parselet {
                     if main {
                         // In case no input was consumed in main loop, skip character
                         if state.is_none()
-                            && context.runtime.reader.capture_from(&reader_start).len() == 0
+                            && context
+                                .runtime
+                                .reader
+                                .capture_from(&context.reader_start)
+                                .len()
+                                == 0
                         {
                             context.runtime.reader.next();
                         }
 
+                        // Update absolute start offset
+                        context.runtime.start = context.runtime.reader.tell().offset;
+
                         // Clear input buffer
                         context.runtime.reader.commit();
 
@@ -197,7 +211,13 @@ impl Parselet {
 
                 Err(reject) => {
                     match reject {
-                        Reject::Skip => break Some(Ok(Accept::Next)),
+                        Reject::Skip => {
+                            if main && state.is_none() {
+                                continue;
+                            }
+
+                            break Some(Ok(Accept::Next));
+                        }
                         Reject::Error(mut err) => {
                             // Patch source position on error, when no position already set
                             if let Some(source_offset) = context.source_offset {
@@ -231,8 +251,11 @@ impl Parselet {
                 state = None;
             }
 
-            // Reset capture stack for loop repeat ($0 must be kept alive)
-            context.runtime.stack.truncate(context.capture_start + 1);
+            // Reset capture stack for loop repeat
+            context.runtime.stack.truncate(context.capture_start - 1); // Truncate stack including $0
+            context.runtime.stack.push(Capture::Empty); // re-initialize $0
+            context.reader_start = context.runtime.reader.tell(); // Reset reader
+
             first = false;
         };
 
@@ -241,7 +264,7 @@ impl Parselet {
             _ => {
                 if results.len() > 1 {
                     Ok(Accept::Push(Capture::Value(
-                        Value::List(Box::new(results)).into_refvalue(),
+                        RefValue::from(results),
                         None,
                         self.severity,
                     )))
@@ -298,52 +321,72 @@ impl Parselet {
             // Check for provided argument count bounds first
             // todo: Not executed when *args-catchall is implemented
             if args > self.signature.len() {
-                return Error::new(
-                    None,
-                    format!(
-                        "Too many parameters, {} possible, {} provided",
+                return Err(match self.signature.len() {
+                    0 => format!(
+                        "{}() doesn't accept any arguments ({} given)",
+                        self.name, args
+                    ),
+                    1 => format!(
+                        "{}() takes exactly one argument ({} given)",
+                        self.name, args
+                    ),
+                    _ => format!(
+                        "{}() expected at most {} arguments ({} given)",
+                        self.name,
                         self.signature.len(),
                         args
                     ),
-                )
-                .into_reject();
+                }
+                .into())
+                .into();
             }
 
             // Set remaining parameters to their defaults
             for (i, arg) in (&self.signature[args..]).iter().enumerate() {
+                // args parameters are previously pushed onto the stack.
                 let var = &mut context.runtime.stack[context.stack_start + args + i];
+
                 //println!("{} {:?} {:?}", i, arg, var);
                 if matches!(var, Capture::Empty) {
-                    // Try to fill argument by named arguments dict
+                    // In case the parameter is empty, try to get it from nargs...
                     if let Some(ref mut nargs) = nargs {
                         if let Some(value) = nargs.remove(&arg.0) {
-                            *var = Capture::Value(value.clone(), None, 0);
+                            *var = Capture::Value(value, None, 0);
                             continue;
                         }
                     }
 
+                    // Otherwise, use default value if available.
                     if let Some(addr) = arg.1 {
-                        // fixme: This might leak the immutablestatic value to something mutable...
+                        // fixme: This might leak the immutable static value to something mutable...
                         *var =
                             Capture::Value(context.runtime.program.statics[addr].clone(), None, 0);
                         //println!("{} receives default {:?}", arg.0, var);
                         continue;
                     }
 
-                    return Error::new(None, format!("Parameter '{}' required", arg.0))
-                        .into_reject();
+                    return Error::new(
+                        None,
+                        format!("{}() expected argument '{}'", self.name, arg.0),
+                    )
+                    .into();
                 }
             }
 
             // Check for remaining nargs
             // todo: Not executed when **nargs-catchall is implemented
-            if let Some(nargs) = nargs {
-                if let Some(narg) = nargs.iter().next() {
-                    return Error::new(
-                        None,
-                        format!("Parameter '{}' provided to call but not used", narg.0),
-                    )
-                    .into_reject();
+            if let Some(mut nargs) = nargs {
+                if let Some((name, _)) = nargs.pop() {
+                    return Err(match nargs.len() {
+                        0 => format!("{}() doesn't accept named argument '{}'", self.name, name),
+                        n => format!(
+                            "{}() doesn't accept named arguments ({} given)",
+                            self.name,
+                            n + 1
+                        ),
+                    }
+                    .into())
+                    .into();
                 }
             }
         } else
@@ -356,7 +399,7 @@ impl Parselet {
         for i in 0..self.locals {
             if let Capture::Empty = context.runtime.stack[context.stack_start + i] {
                 context.runtime.stack[context.stack_start + i] =
-                    Capture::Value(Value::Void.into_refvalue(), None, 0);
+                    Capture::Value(crate::value!(void), None, 0);
             }
         }
 
@@ -421,7 +464,7 @@ impl Parselet {
                 context
                     .runtime
                     .stack
-                    .resize(context.capture_start + 1, Capture::Empty);
+                    .resize(context.capture_start, Capture::Empty);
             }
 
             context.runtime.reader.reset(reader_end);
@@ -440,11 +483,14 @@ impl Parselet {
             result
         };
 
+        /*
+        // Dump AST when parselet returns an AST for debugging purposes.
+        // fixme: Disabled for now, can be enabled on demand.
         if context.runtime.debug > 1 {
             loop {
                 if let Ok(Accept::Push(Capture::Value(ref value, ..))) = result {
                     let value = value.borrow();
-                    if let Some(d) = value.get_dict() {
+                    if let Some(d) = value.dict() {
                         if d.get("emit").is_some() {
                             context.debug("=> AST");
                             ast::print(&value);
@@ -457,30 +503,210 @@ impl Parselet {
                 break;
             }
         }
+        */
 
         result
     }
 }
 
-impl std::cmp::PartialEq for Parselet {
-    // It satisfies to just compare the parselet's memory address for equality
+impl From<Parselet> for RefValue {
+    fn from(parselet: Parselet) -> Self {
+        RefValue::from(Box::new(ParseletRef(Rc::new(RefCell::new(parselet)))) as BoxedObject)
+    }
+}
+
+#[derive(Clone, Debug)]
+pub struct ParseletRef(pub Rc<RefCell<Parselet>>);
+
+impl Object for ParseletRef {
+    fn id(&self) -> usize {
+        &*self.0.borrow() as *const Parselet as usize
+    }
+
+    fn name(&self) -> &'static str {
+        "parselet"
+    }
+
+    fn is_callable(&self, without_arguments: bool) -> bool {
+        let parselet = self.0.borrow();
+
+        if without_arguments {
+            parselet.signature.len() == 0 || parselet.signature.iter().all(|arg| arg.1.is_some())
+        } else {
+            true
+        }
+    }
+
+    fn is_consuming(&self) -> bool {
+        self.0.borrow().consuming.is_some()
+    }
+
+    fn call(
+        &self,
+        context: &mut Context,
+        args: usize,
+        nargs: Option<Dict>,
+    ) -> Result<Accept, Reject> {
+        self.0
+            .borrow()
+            .run(context.runtime, args, nargs, false, context.depth + 1)
+    }
+}
+
+impl PartialEq for ParseletRef {
     fn eq(&self, other: &Self) -> bool {
-        self as *const Parselet as usize == other as *const Parselet as usize
+        self.id() == other.id()
     }
 }
 
-impl std::hash::Hash for Parselet {
-    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
-        (self as *const Parselet as usize).hash(state);
-    }
-}
-
-impl std::cmp::PartialOrd for Parselet {
-    // It satisfies to just compare the parselet's memory address for equality
+impl PartialOrd for ParseletRef {
     fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
-        let left = self as *const Parselet as usize;
-        let right = other as *const Parselet as usize;
-
-        left.partial_cmp(&right)
+        self.id().partial_cmp(&other.id())
     }
 }
+
+#[test]
+fn test_function_as_static_with_args() {
+    assert_eq!(
+        crate::run(
+            "
+            faculty : @x {
+                if !x return 1
+                x * faculty(x - 1)
+            }
+
+            faculty(4)
+            ",
+            ""
+        ),
+        Ok(Some(crate::value!(24)))
+    );
+}
+
+#[test]
+fn test_function_as_variable_with_args() {
+    assert_eq!(
+        crate::run(
+            "
+            faculty = @x {
+                if !x return 1
+                x * faculty(x - 1)
+            }
+
+            faculty(4)
+            ",
+            ""
+        ),
+        Ok(Some(crate::value!(24)))
+    );
+}
+
+#[test]
+fn test_parselet_call_error_reporting() {
+    // Tests for calling single argument parselet
+    for (call, msg) in [
+        ("f()", "Line 2, column 1: f() expected argument 'x'"),
+        (
+            "f(1, 2)",
+            "Line 2, column 1: f() takes exactly one argument (2 given)",
+        ),
+        (
+            "f(1, y=2)",
+            "Line 2, column 1: f() doesn't accept named argument 'y'",
+        ),
+    ] {
+        let call = format!("f : @x {{ x * x }}\n{}", call);
+
+        assert_eq!(crate::run(&call, ""), Err(msg.to_owned()));
+    }
+
+    // Tests for calling mutli-argument parselet with wrong arguments counts
+    for (call, msg) in [
+        ("foo()", "Line 2, column 1: Call to unresolved symbol 'foo'"),
+        ("f()", "Line 2, column 1: f() expected argument 'a'"),
+        (
+            "f(1, 2, 3, 4)",
+            "Line 2, column 1: f() expected at most 3 arguments (4 given)",
+        ),
+        (
+            "f(c=10, d=3, e=10)",
+            "Line 2, column 1: f() expected argument 'a'",
+        ),
+        (
+            "f(1, c=10, d=3)",
+            "Line 2, column 1: f() doesn't accept named argument 'd'",
+        ),
+        (
+            "f(1, c=10, d=3, e=7)",
+            "Line 2, column 1: f() doesn't accept named arguments (2 given)",
+        ),
+    ] {
+        let call = format!("f : @a, b=2, c {{ a b c }}\n{}", call);
+
+        assert_eq!(crate::run(&call, ""), Err(msg.to_owned()));
+    }
+}
+
+#[test]
+fn test_parselet_begin_end() {
+    assert_eq!(
+        crate::run(
+            "
+            begin { x = 0 1337 }
+            end 1338
+
+            P: @{ 'lol' x = x + 1 x }
+            P",
+            "lolalolaalolol"
+        ),
+        Ok(Some(crate::value!([1337, 1, 2, 3, 1338])))
+    );
+
+    assert_eq!(
+        crate::run(
+            "
+            begin x = 1
+
+            'lol' $1 * x x x = x + 1",
+            "lolAlolBlol"
+        ),
+        Ok(Some(crate::value!([
+            ["lol", 1],
+            ["lollol", 2],
+            ["lollollol", 3]
+        ])))
+    );
+
+    // begin and end without any input
+    assert_eq!(
+        crate::run(
+            "
+            begin 1
+            2 3 4
+            end 5
+            ",
+            ""
+        ),
+        Ok(Some(crate::value!([1, [2, 3, 4], 5])))
+    )
+}
+
+#[test]
+fn test_parselet_leftrec() {
+    assert_eq!(
+        crate::run("P: @{ P? ''a'' }\nP", "aaaa"),
+        Ok(Some(crate::value!([[["a", "a"], "a"], "a"])))
+    );
+
+    // todo: More examples here please!
+}
+
+#[test]
+fn test_parselet_repeat() {
+    assert_eq!(
+        crate::run("P: @{ 'a' repeat $1 }\nP", "aaaa"),
+        Ok(Some(crate::value!(["a", "a", "a", "a"])))
+    );
+
+    // todo: More examples here please!
+}
