@@ -1,10 +1,11 @@
-use std::io;
-use std::io::prelude::*;
-
 use super::*;
 use crate::error::Error;
 use crate::reader::Offset;
-use crate::value::{Dict, Value};
+use crate::value;
+use crate::value::{Dict, Object, Str, Value};
+use std::io;
+use std::io::prelude::*;
+use std::rc::Rc;
 
 // --- Op ----------------------------------------------------------------------
 
@@ -21,12 +22,12 @@ pub enum Op {
     Rust(Rust),          // Native rust callback
 
     // Capture frames
-    Frame(usize),   // Start new frame with optional forward fuse
-    Commit,         // Commit frame
-    Reset,          // Reset frame
-    Close,          // Close frame
-    Collect(usize), // Collect stack values from current frame
-    Fuse(usize),    // Set frame fuse to forward address
+    Frame(usize),    // Start new frame with optional forward fuse
+    Commit,          // Commit frame
+    Reset,           // Reset frame
+    Close,           // Close frame
+    Collect(u8, u8), // Collect stack values from current frame
+    Fuse(usize),     // Set frame fuse to forward address
 
     // Loop frames
     Loop(usize), // Loop frame
@@ -94,38 +95,21 @@ pub enum Op {
     StoreFastCaptureHold(usize),
     StoreCapture,
     StoreCaptureHold,
+    StoreIndex,
+    StoreIndexHold,
 
     MakeAlias,       // Make key-value-Capture from last two stack items
     MakeDict(usize), // Make a Dict from specified amount of key-value-pairs
 
     // Operations
     Drop,  // drop TOS
+    Sep,   // separate, ensure TOS value is not shared
     Clone, // clone TOS
     Dup,   // duplicate TOS
     Rot2,  // rotate TOS by 2
 
-    Add, // binary add
-    Sub, // binary sub
-    Mul, // binary mul
-    Div, // binary div
-
-    Not, // unary not (! operator)
-    Neg, // unary negation (- operator)
-
-    InlineAdd, // Inline add (+= operator)
-    InlineSub, // Inline sub (-= operator)
-    InlineMul, // Inline mul (*= operator)
-    InlineDiv, // Inline div (/= operator)
-
-    InlineInc, // Inline increment (++x and x++ operators)
-    InlineDec, // Inline decrement (--x and x-- operators)
-
-    Equal,        // Compare for equality (== operator)
-    NotEqual,     // Compare for unequality (!= operator)
-    LowerEqual,   // Compare for lower-equality (<= operator)
-    GreaterEqual, // Compare for greater-equality (>= operator)
-    Lower,        // Compare for lowerness (< operator)
-    Greater,      // Compare for greaterness (> operator)
+    UnaryOp(&'static str),  // Operation with one operand
+    BinaryOp(&'static str), // Operation with two operands
 }
 
 impl Op {
@@ -245,10 +229,10 @@ impl Op {
                     Ok(Accept::Next)
                 }
 
-                Op::Collect(severity) => {
-                    match context.collect(frame.capture_start, false, true, true, *severity as u8) {
+                Op::Collect(collect, push) => {
+                    match context.collect(frame.capture_start, false, true, true, *collect) {
                         Err(capture) => Ok(Accept::Push(capture)),
-                        Ok(Some(value)) => Ok(Accept::Push(Capture::Value(value, None, 5))),
+                        Ok(Some(value)) => Ok(Accept::Push(Capture::Value(value, None, *push))),
                         Ok(None) => Ok(Accept::Next),
                     }
                 }
@@ -309,7 +293,7 @@ impl Op {
 
                 // Conditional jumps
                 Op::ForwardIfTrue(goto) => {
-                    if context.pop().borrow().is_true() {
+                    if context.pop().is_true() {
                         ip += goto;
                     } else {
                         ip += 1;
@@ -319,7 +303,7 @@ impl Op {
                 }
 
                 Op::ForwardIfFalse(goto) => {
-                    if !context.pop().borrow().is_true() {
+                    if !context.pop().is_true() {
                         ip += goto;
                     } else {
                         ip += 1;
@@ -338,7 +322,7 @@ impl Op {
                 }
 
                 Op::BackwardIfTrue(goto) => {
-                    if context.pop().borrow().is_true() {
+                    if context.pop().is_true() {
                         ip -= goto;
                     } else {
                         ip += 1;
@@ -348,7 +332,7 @@ impl Op {
                 }
 
                 Op::BackwardIfFalse(goto) => {
-                    if !context.pop().borrow().is_true() {
+                    if !context.pop().is_true() {
                         ip -= goto;
                     } else {
                         ip += 1;
@@ -398,97 +382,92 @@ impl Op {
                 }
                 Op::Reject => Err(Reject::Return),
                 Op::LoadExit => {
-                    std::process::exit(context.pop().borrow().to_integer() as i32);
+                    std::process::exit(context.pop().to_i64() as i32);
                 }
                 Op::Exit => std::process::exit(0),
 
                 Op::Error(msg) => {
                     if let Some(msg) = msg {
-                        Error::new(Some(frame.reader_start), msg.clone()).into_reject()
+                        Error::new(Some(frame.reader_start), msg.clone()).into()
                     } else {
-                        Error::new(Some(frame.reader_start), context.pop().borrow().to_string())
-                            .into_reject()
+                        Error::new(Some(frame.reader_start), context.pop().to_string()).into()
                     }
                 }
 
                 // Calls
                 Op::CallOrCopy => {
                     let value = context.pop();
-                    let value = value.borrow();
 
-                    if value.is_callable(false) {
+                    if value.is_callable(true) {
                         // Call the value without parameters
                         value.call(context, 0, None)
                     } else {
                         // Push a copy of the value
-                        context.push(value.clone().into_refvalue())
+                        context.push(value.borrow().clone().into())
                     }
                 }
 
                 Op::Call => {
                     let target = context.pop();
-                    let target = target.borrow();
                     target.call(context, 0, None)
                 }
 
                 Op::CallArg(args) => {
                     let target = context.pop();
-                    let target = target.borrow();
                     target.call(context, *args, None)
                 }
 
                 Op::CallArgNamed(args) => {
                     let target = context.pop();
-                    let target = target.borrow();
+                    let nargs = Value::from(context.pop());
 
-                    let nargs = Value::from_ref(context.pop()).unwrap();
-                    target.call(context, *args, Some(nargs.into_dict()))
-                    //println!("CallArgNamed returns {:?}", ret);
+                    if let Some(nargs) = nargs.into_object::<Dict>() {
+                        target.call(context, *args, Some(nargs))
+                    } else {
+                        panic!("nargs operand required to be dict")
+                    }
                 }
 
-                Op::CallStatic(addr) => context.runtime.program.statics[*addr]
-                    .borrow()
-                    .call(context, 0, None),
+                Op::CallStatic(addr) => {
+                    context.runtime.program.statics[*addr].call(context, 0, None)
+                }
 
                 Op::CallStaticArg(addr_args) => {
-                    context.runtime.program.statics[addr_args.0].borrow().call(
-                        context,
-                        addr_args.1,
-                        None,
-                    )
+                    context.runtime.program.statics[addr_args.0].call(context, addr_args.1, None)
                     //println!("CallStaticArg returns {:?}", ret);
                 }
 
                 Op::CallStaticArgNamed(addr_args) => {
-                    let nargs = Value::from_ref(context.pop()).unwrap();
+                    let nargs = Value::from(context.pop());
 
-                    context.runtime.program.statics[addr_args.0].borrow().call(
-                        context,
-                        addr_args.1,
-                        Some(nargs.into_dict()),
-                    )
-                    //println!("CallStaticArg returns {:?}",
+                    if let Some(nargs) = nargs.into_object::<Dict>() {
+                        context.runtime.program.statics[addr_args.0].call(
+                            context,
+                            addr_args.1,
+                            Some(nargs),
+                        )
+                    } else {
+                        panic!("nargs operand required to be dict")
+                    }
                 }
 
                 // Variables and values
                 Op::LoadStatic(addr) => {
                     let value = &context.runtime.program.statics[*addr];
-                    context.push(value.borrow().clone().into_refvalue())
+                    context.push(value.borrow().clone().into())
                 }
-                Op::Push0 => context.push(Value::Integer(0).into_refvalue()),
-                Op::Push1 => context.push(Value::Integer(1).into_refvalue()),
-                Op::PushVoid => context.push(Value::Void.into_refvalue()),
-                Op::PushNull => context.push(Value::Null.into_refvalue()),
-                Op::PushTrue => context.push(Value::True.into_refvalue()),
-                Op::PushFalse => context.push(Value::False.into_refvalue()),
+                Op::Push0 => context.push(value!(0 as i64)),
+                Op::Push1 => context.push(value!(1 as i64)),
+                Op::PushVoid => context.push(value!(void)),
+                Op::PushNull => context.push(value!(null)),
+                Op::PushTrue => context.push(value!(true)),
+                Op::PushFalse => context.push(value!(false)),
 
                 Op::LoadGlobal(addr) => context.load(*addr),
                 Op::LoadFast(addr) => context.load(context.stack_start + *addr),
 
                 Op::LoadFastCapture(index) => {
-                    let value = context
-                        .get_capture(*index)
-                        .unwrap_or(Value::Void.into_refvalue());
+                    let value = context.get_capture(*index).unwrap_or(value!(void));
                     context.push(value)
                 }
 
@@ -496,23 +475,17 @@ impl Op {
                     let index = context.pop();
                     let index = index.borrow();
 
-                    match &*index {
-                        Value::Addr(_) | Value::Integer(_) | Value::Float(_) => {
-                            let value = context
-                                .get_capture(index.to_addr())
-                                .unwrap_or(Value::Void.into_refvalue());
-                            context.push(value)
-                        }
+                    let value = if let Some(alias) = index.object::<Str>() {
+                        context
+                            .get_capture_by_name(alias.as_str())
+                            .unwrap_or(value!(void))
+                    } else {
+                        context
+                            .get_capture(index.to_usize())
+                            .unwrap_or(value!(void))
+                    };
 
-                        Value::String(alias) => {
-                            let value = context
-                                .get_capture_by_name(alias)
-                                .unwrap_or(Value::Void.into_refvalue());
-                            context.push(value)
-                        }
-
-                        _ => Ok(Accept::Next),
-                    }
+                    context.push(value)
                 }
 
                 Op::LoadAttr => {
@@ -520,13 +493,15 @@ impl Op {
                     let attr = attr.borrow();
                     let value = context.pop();
 
-                    match Value::get_attr(value, &attr) {
+                    match value.create_method(attr.object::<Str>().unwrap().as_str()) {
                         Ok(value) => context.push(value),
-                        Err(msg) => Error::new(None, msg).into_reject(),
+                        Err(msg) => Error::new(None, msg).into(),
                     }
                 }
 
                 Op::LoadIndex => {
+                    //fixme
+                    /*
                     let index = context.pop();
                     let index = index.borrow();
                     let value = context.pop();
@@ -534,8 +509,10 @@ impl Op {
 
                     match value.get_index(&index) {
                         Ok(value) => context.push(value),
-                        Err(msg) => Error::new(None, msg).into_reject(),
+                        Err(msg) => Error::new(None, msg).into(),
                     }
+                    */
+                    todo!();
                 }
 
                 Op::StoreGlobal(addr) => {
@@ -549,7 +526,7 @@ impl Op {
                     // todo: bounds checking?
                     let value = context.peek();
                     context.runtime.stack[*addr] =
-                        Capture::Value(value.borrow().clone().into_refvalue(), None, 0);
+                        Capture::Value(value.borrow().clone().into(), None, 0);
                     Ok(Accept::Next)
                 }
 
@@ -565,7 +542,7 @@ impl Op {
                     // todo: bounds checking?
                     let value = context.peek();
                     context.runtime.stack[context.stack_start + *addr] =
-                        Capture::Value(value.borrow().clone().into_refvalue(), None, 0);
+                        Capture::Value(value.borrow().clone().into(), None, 0);
                     Ok(Accept::Next)
                 }
 
@@ -579,7 +556,7 @@ impl Op {
                 Op::StoreFastCaptureHold(index) => {
                     let value = context.peek();
 
-                    context.set_capture(*index, value.borrow().clone().into_refvalue());
+                    context.set_capture(*index, value.borrow().clone().into());
                     Ok(Accept::Next)
                 }
 
@@ -587,44 +564,54 @@ impl Op {
                     let index = context.pop();
                     let index = index.borrow();
 
-                    match &*index {
-                        Value::Addr(_) | Value::Integer(_) | Value::Float(_) => {
-                            if matches!(op, Op::StoreCapture) {
-                                let value = context.pop();
-                                context.set_capture(index.to_addr(), value);
-                                Ok(Accept::Push(Capture::Empty))
-                            } else {
-                                let value = context.peek();
-                                context.set_capture(
-                                    index.to_addr(),
-                                    value.borrow().clone().into_refvalue(),
-                                );
-                                Ok(Accept::Next)
-                            }
+                    if let Some(alias) = index.object::<Str>() {
+                        if matches!(op, Op::StoreCapture) {
+                            let value = context.pop();
+                            context.set_capture_by_name(alias, value);
+                            Ok(Accept::Push(Capture::Empty))
+                        } else {
+                            let value = context.peek();
+                            context.set_capture_by_name(alias, value.borrow().clone().into());
+                            Ok(Accept::Next)
                         }
-
-                        Value::String(alias) => {
-                            if matches!(op, Op::StoreCapture) {
-                                let value = context.pop();
-                                context.set_capture_by_name(alias, value);
-                                Ok(Accept::Push(Capture::Empty))
-                            } else {
-                                let value = context.peek();
-                                context.set_capture_by_name(
-                                    alias,
-                                    value.borrow().clone().into_refvalue(),
-                                );
-                                Ok(Accept::Next)
-                            }
+                    } else {
+                        if matches!(op, Op::StoreCapture) {
+                            let value = context.pop();
+                            context.set_capture(index.to_usize(), value);
+                            Ok(Accept::Push(Capture::Empty))
+                        } else {
+                            let value = context.peek();
+                            context.set_capture(index.to_usize(), value.borrow().clone().into());
+                            Ok(Accept::Next)
                         }
-
-                        _ => unimplemented!(),
                     }
                 }
 
+                Op::StoreIndex | Op::StoreIndexHold => {
+                    //fixme
+                    /*
+                    let index = context.pop();
+                    let index = index.borrow();
+                    let target = context.pop();
+                    let value = context.pop();
+
+                    let mut obj = target.borrow_mut();
+
+                    if let Err(msg) = obj.set_index(&index, value) {
+                        Error::new(None, msg).args[0].as_ref().unwrap().()
+                    } else {
+                        if matches!(op, Op::StoreIndexHold) {
+                            context.push(target.clone())
+                        } else {
+                            Ok(Accept::Next)
+                        }
+                    }
+                    */
+                    todo!();
+                }
+
                 Op::MakeAlias => {
                     let name = context.pop();
-                    let name = name.borrow();
 
                     match context.runtime.stack.last_mut().unwrap() {
                         Capture::Range(_, alias, ..) | Capture::Value(_, alias, ..) => {
@@ -632,11 +619,7 @@ impl Op {
                         }
 
                         empty => {
-                            *empty = Capture::Value(
-                                Value::Void.into_refvalue(),
-                                Some(name.to_string()),
-                                0,
-                            );
+                            *empty = Capture::Value(value!(void), Some(name.to_string()), 0);
                         }
                     }
 
@@ -648,13 +631,12 @@ impl Op {
 
                     for _ in 0..*count {
                         let key = context.pop();
-                        let key = key.borrow();
 
                         let value = context.pop();
                         dict.insert(key.to_string(), value);
                     }
 
-                    context.push(Value::Dict(Box::new(dict)).into_refvalue())
+                    context.push(RefValue::from(dict))
                 }
 
                 // Operations
@@ -663,6 +645,19 @@ impl Op {
                     Ok(Accept::Next)
                 }
 
+                Op::Sep => {
+                    let mut value = context.pop();
+
+                    if Rc::strong_count(&value) > 1 {
+                        value = RefValue::from({
+                            let inner = value.borrow();
+                            inner.clone()
+                        });
+                    }
+
+                    context.push(value)
+                }
+
                 Op::Clone => {
                     let value = context.peek();
                     context.push(value.clone())
@@ -671,7 +666,7 @@ impl Op {
                 Op::Dup => {
                     let value = context.peek();
                     let value = value.borrow();
-                    context.push(value.clone().into_refvalue())
+                    context.push(value.clone().into())
                 }
 
                 Op::Rot2 => {
@@ -684,101 +679,15 @@ impl Op {
                     Ok(Accept::Next)
                 }
 
-                // Operations
-                Op::Add | Op::Sub | Op::Mul | Op::Div => {
-                    let b = context.pop();
-                    let a = context.pop();
-
-                    /*
-                    println!("{:?}", op);
-                    println!("a = {:?}", a);
-                    println!("b = {:?}", b);
-                    */
-
-                    let c = match op {
-                        Op::Add => a.borrow().add(&*b.borrow())?.into_refvalue(),
-                        Op::Sub => a.borrow().sub(&*b.borrow())?.into_refvalue(),
-                        Op::Mul => a.borrow().mul(&*b.borrow())?.into_refvalue(),
-                        Op::Div => a.borrow().div(&*b.borrow())?.into_refvalue(),
-                        _ => unimplemented!("Unimplemented operator"),
-                    };
-
-                    context.push(c)
-                }
-
-                Op::Equal
-                | Op::NotEqual
-                | Op::LowerEqual
-                | Op::GreaterEqual
-                | Op::Lower
-                | Op::Greater => {
-                    let b = context.pop();
-                    let a = context.pop();
-
-                    //println!("{:?}", op);
-                    //println!("a = {:?}", a);
-                    //println!("b = {:?}", b);
-
-                    let c = match op {
-                        Op::Equal => &*a.borrow() == &*b.borrow(),
-                        Op::NotEqual => &*a.borrow() != &*b.borrow(),
-                        Op::LowerEqual => &*a.borrow() <= &*b.borrow(),
-                        Op::GreaterEqual => &*a.borrow() >= &*b.borrow(),
-                        Op::Lower => &*a.borrow() < &*b.borrow(),
-                        Op::Greater => &*a.borrow() > &*b.borrow(),
-
-                        _ => unimplemented!("Unimplemented operator"),
-                    };
-
-                    //println!("c = {:?}", c);
-
-                    context.push((if c { Value::True } else { Value::False }).into_refvalue())
-                }
-
-                Op::Not => {
-                    let value = context.pop().borrow().not()?.into_refvalue();
-                    context.push(value)
-                }
-                Op::Neg => {
-                    let value = context.pop().borrow().neg()?.into_refvalue();
-                    context.push(value)
-                }
-                Op::InlineAdd | Op::InlineSub | Op::InlineMul | Op::InlineDiv => {
-                    let b = context.pop();
+                Op::UnaryOp(op) => {
                     let value = context.pop();
-                    let mut value = value.borrow_mut();
-
-                    /*
-                    println!("{:?}", op);
-                    println!("a = {:?}", a);
-                    println!("b = {:?}", b);
-                    */
-
-                    *value = match op {
-                        Op::InlineAdd => value.add(&*b.borrow())?,
-                        Op::InlineSub => value.sub(&*b.borrow())?,
-                        Op::InlineMul => value.mul(&*b.borrow())?,
-                        Op::InlineDiv => value.div(&*b.borrow())?,
-                        _ => unimplemented!("Unimplemented operator"),
-                    };
-
-                    context.push(value.clone().into_refvalue())
+                    context.push(value.unary_op(op)?)
                 }
 
-                Op::InlineInc => {
-                    let value = context.pop();
-                    let mut value = value.borrow_mut();
-
-                    *value = value.add(&Value::Integer(1))?; // todo: perform inc by bit-shift
-                    context.push(value.clone().into_refvalue())
-                }
-
-                Op::InlineDec => {
-                    let value = context.pop();
-                    let mut value = value.borrow_mut();
-
-                    *value = value.sub(&Value::Integer(1))?; // todo: perform dec by bit-shift
-                    context.push(value.clone().into_refvalue())
+                Op::BinaryOp(op) => {
+                    let last = context.pop();
+                    let first = context.pop();
+                    context.push(first.binary_op(last, op)?)
                 }
             };
 
@@ -818,19 +727,6 @@ impl Op {
             }
         }
 
-        // Take last remaining value as result (if some)
-        if let Ok(_) = state {
-            state = match context.runtime.stack.len() - frame.capture_start {
-                0 => Ok(Accept::Next),
-                _ => Ok(Accept::Push(context.runtime.stack.pop().unwrap())),
-            };
-        }
-
-        // Debug
-        if context.runtime.debug > 2 {
-            context.debug(&format!("returns {:?}", state));
-        }
-
         state
     }
 }
